# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy.signal

from ..signal import signal_smooth




def ecg_findpeaks(ecg_cleaned, sampling_rate=1000, method="neurokit", show=False):
    """Find R-peaks in an ECG signal.

    Find R-peaks in an ECG signal using the specified method.

    Parameters
    ----------
    ecg_cleaned : list, array or Series
        The cleaned ECG channel as returned by `ecg_clean()`.
    sampling_rate : int
        The sampling frequency of `ecg_signal` (in Hz, i.e., samples/second).
        Defaults to 1000.
    method : string
        The algorithm to be used for R-peak detection. Defaults to "neurokit".
    show : bool
        If True, will return a plot to visualizing the thresholds used in the
        algorithm. Useful for debugging.

    Returns
    -------
    signals : DataFrame
        A DataFrame of same length as the input signal in which occurences of
        R-peaks marked as "1" in a list of zeros with the same length as
        `ecg_cleaned`. Accessible with the keys "ECG_Peaks".
    info : dict
        A dictionary containing additional information, in this case the
        samples at which R-peaks occur, accessible with the key "ECG_Peaks".

    See Also
    --------
    ecg_clean, ecg_rate, ecg_process, ecg_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> ecg = nk.ecg_simulate(duration=10, sampling_rate=1000)
    >>> cleaned = nk.ecg_clean(ecg, sampling_rate=1000)
    >>> signals, info = nk.ecg_findpeaks(cleaned)
    >>> nk.events_plot(info["ECG_Peaks"], cleaned)
    """
    # Try retrieving right column
    if isinstance(ecg_cleaned, pd.DataFrame):
        try:
            ecg_cleaned = ecg_cleaned["RSP_Clean"]
        except NameError:
            try:
                ecg_cleaned = ecg_cleaned["ECG_Raw"]
            except NameError:
                ecg_cleaned = ecg_cleaned["RSP"]

    cleaned = np.array(ecg_cleaned)

    method = method.lower()  # remove capitalised letters

    # Run peak detection algorithm
    if method == "neurokit":
        peaks = _ecg_findpeaks_neurokit(cleaned,
                                        sampling_rate,
                                        smoothwindow=.1,
                                        avgwindow=.75,
                                        gradthreshweight=1.5,
                                        minlenweight=0.4,
                                        mindelay=0.3,
                                        show=show)
    elif method == "pamtompkins":
        peaks = _ecg_findpeaks_pantompkins()
    else:
        raise ValueError("NeuroKit error: ecg_findpeaks(): 'method' should be "
                         "one of 'neurokit' or 'pamtompkins'.")


    # Prepare output.
    peaks_signal = np.zeros(len(cleaned))
    peaks_signal[peaks] = 1
    signals = pd.DataFrame({"ECG_Peaks": peaks_signal})

    info = {"ECG_Peaks": peaks}

    return signals, info



# =============================================================================
# Pantompkins
# =============================================================================
def _ecg_findpeaks_pantompkins():
    raise ValueError("NeuroKit error: ecg_findpeaks(): pamtompkins 'method' "
                     "is not implemented yet.")



# =============================================================================
# NeuroKit
# =============================================================================
def _ecg_findpeaks_neurokit(signal, sampling_rate, smoothwindow=.1, avgwindow=.75,
                            gradthreshweight=1.5, minlenweight=0.4, mindelay=0.3,
                            show=False):
    """
    All tune-able parameters are specified as keyword arguments. The `signal`
    must be the highpass-filtered raw ECG with a lowcut of .5 Hz.
    """
    if show is True:
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True)

    # Compute the ECG's gradient as well as the gradient threshold. Run with
    # show=True in order to get an idea of the threshold.
    grad = np.gradient(signal)
    absgrad = np.abs(grad)
    smooth_kernel = int(np.rint(smoothwindow * sampling_rate))
    avg_kernel = int(np.rint(avgwindow * sampling_rate))
    smoothgrad = signal_smooth(absgrad, kernel="boxcar", size=smooth_kernel)
    avggrad = signal_smooth(smoothgrad, kernel="boxcar", size=avg_kernel)
    gradthreshold = gradthreshweight * avggrad
    mindelay = int(np.rint(sampling_rate * mindelay))

    if show is True:
        ax1.plot(signal)
        ax2.plot(smoothgrad)
        ax2.plot(gradthreshold)

    # Identify start and end of QRS complexes.
    qrs = smoothgrad > gradthreshold
    beg_qrs = np.where(np.logical_and(np.logical_not(qrs[0:-1]), qrs[1:]))[0]
    end_qrs = np.where(np.logical_and(qrs[0:-1], np.logical_not(qrs[1:])))[0]
    # Throw out QRS-ends that precede first QRS-start.
    end_qrs = end_qrs[end_qrs > beg_qrs[0]]

    # Identify R-peaks within QRS (ignore QRS that are too short).
    num_qrs = min(beg_qrs.size, end_qrs.size)
    min_len = np.mean(end_qrs[:num_qrs] - beg_qrs[:num_qrs]) * minlenweight
    peaks = [0]

    for i in range(num_qrs):

        beg = beg_qrs[i]
        end = end_qrs[i]
        len_qrs = end - beg

        if len_qrs < min_len:
            continue

        if show is True:
            ax2.axvspan(beg, end, facecolor="m", alpha=0.5)

        # Find local maxima and their prominence within QRS.
        data = signal[beg:end]
        locmax, props = scipy.signal.find_peaks(data, prominence=(None, None))

        if locmax.size > 0:
            # Identify most prominent local maximum.
            peak = beg + locmax[np.argmax(props["prominences"])]
            # Enforce minimum delay between peaks.
            if peak - peaks[-1] > mindelay:
                peaks.append(peak)

    peaks.pop(0)

    if show is True:
        ax1.scatter(peaks, signal[peaks], c="r")

    peaks = np.asarray(peaks).astype(int)  # Convert to int
    return peaks



# -*- coding: utf-8 -*-
import numpy as np
import mne

import scipy.signal



def signal_filter(signal, sampling_rate=1000, lowcut=None, highcut=None, method="butterworth", order=2, window_length="default"):
    """Filter a signal using 'butterworth', 'fir' or 'savgol' filters.

    Apply a lowpass (if 'highcut' frequency is provided), highpass (if 'lowcut' frequency is provided) or bandpass (if both are provided) filter to the signal.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
        or "bandstop".
    sampling_rate : int
        The sampling frequency of the signal (in Hz, i.e., samples/second).
    lowcut : float
        Lower cutoff frequency in Hz. The default is None.
    highcut : float
        Upper cutoff frequency in Hz. The default is None.
    method : str
        Can be one of 'butterworth', 'fir' or 'savgol'.
    order : int
        Only used if method is 'butterworth' or 'savgol'. Order of the filter (default is 2).
    window_length : int
        Only used if method is 'savgol'. The length of the filter window (i.e. the number of coefficients). Must be an odd integer. If 'default', will be set to the sampling rate divided by 10 (101 if the sampling rate is 1000 Hz).

    See Also
    --------
    signal_detrend, signal_psd

    Returns
    -------
    array
        Vector containing the filtered signal.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=50, num=10000)) # Low freq
    >>> signal += np.cos(np.linspace(start=0, stop=500, num=10000)) # High freq
    >>>
    >>> # Lowpass
    >>> pd.DataFrame({"Raw": signal,
                      "Butter_2": nk.signal_filter(signal, highcut=3, method='butterworth', order=2),
                      "Butter_3": nk.signal_filter(signal, highcut=3, method='butterworth', order=3),
                      "Butter_5": nk.signal_filter(signal, highcut=3, method='butterworth', order=5),
                      "FIR": nk.signal_filter(signal, highcut=3, method='fir')}).plot(subplots=True)

    >>># Highpass
    >>> pd.DataFrame({"Raw": signal,
                      "Butter_2": nk.signal_filter(signal, lowcut=2, method='butterworth', order=2),
                      "Butter_3": nk.signal_filter(signal, lowcut=2, method='butterworth', order=3),
                      "Butter_5": nk.signal_filter(signal, lowcut=2, method='butterworth', order=5),
                      "FIR": nk.signal_filter(signal, lowcut=2, method='fir')}).plot(subplots=True)

    >>> original = nk.rsp_simulate(duration=30, method="sinusoidal")
    >>> signal = original + np.cos(np.linspace(start=0, stop=15, num=len(original)))/2  # Low freq
    >>> signal += np.cos(np.linspace(start=0, stop=1000, num=len(signal)))/2  # High freq
    >>>
    >>> pd.DataFrame({"Raw": signal,
                      "Lowpass": nk.signal_filter(signal, highcut=2),
                      "Highpass": nk.signal_filter(signal, lowcut=0.1),
                      "Bandpass": nk.signal_filter(signal, lowcut=0.1, highcut=2),
                      "Savgol": nk.signal_filter(signal, method='savgol')}).plot(subplots=True)
    """
    # Sanity checks
    if lowcut is None and highcut is None:
        return signal

    method = method.lower()
    if method in ["sg", "savgol", "savitzky-golay"]:
        filtered = _signal_filter_savgol(signal, sampling_rate, order, window_length=window_length)
    elif method in ["butter", "butterworth"]:
        filtered = _signal_filter_butterworth(signal, sampling_rate, lowcut, highcut, order)
    elif method in ["fir"]:
        filtered = _signal_filter_fir(signal, sampling_rate, lowcut, highcut, window_length=window_length)
    else:
        raise ValueError("NeuroKit error: signal_filter(): 'method' should be "
                         "one of 'butterworth', 'savgol' or 'fir'.")
    return filtered


# =============================================================================
# Savitzky-Golay (savgol)
# =============================================================================

def _signal_filter_savgol(signal, sampling_rate=1000, order=2, window_length="default"):
    """Filter a signal using the Savitzky-Golay method.

    Default window size is chosen based on `Sadeghi, M., & Behnia, F. (2018). Optimum window length of Savitzky-Golay filters with arbitrary order. arXiv preprint arXiv:1808.10489. <https://arxiv.org/ftp/arxiv/papers/1808/1808.10489.pdf>`_.
    """
    window_length = _signal_filter_windowlength(window_length=window_length, sampling_rate=sampling_rate)

    filtered = scipy.signal.savgol_filter(signal, window_length=window_length, polyorder=order)
    return filtered

# =============================================================================
# FIR
# =============================================================================
def _signal_filter_fir(signal, sampling_rate=1000, lowcut=None, highcut=None, window_length="default"):
    """Filter a signal using a FIR filter.
    """
    if isinstance(window_length, str):
        window_length = "auto"

    filtered = mne.filter.filter_data(
            signal,
            sfreq=sampling_rate,
            l_freq=lowcut,
            h_freq=highcut,
            method='fir',
            fir_window='hamming',
            filter_length=window_length,
            l_trans_bandwidth='auto',
            h_trans_bandwidth='auto',
            phase='zero-double',
            fir_design='firwin',
            pad='reflect_limited',
            verbose=False)
    return filtered


# =============================================================================
# Butterworth
# =============================================================================

def _signal_filter_butterworth(signal, sampling_rate=1000, lowcut=None, highcut=None, order=5):
    """Filter a signal using IIR Butterworth SOS method.
    """
    freqs, filter_type = _signal_filter_sanitize(lowcut=lowcut, highcut=highcut)

    sos = scipy.signal.butter(order, freqs, btype=filter_type, output='sos', fs=sampling_rate)
    filtered = scipy.signal.sosfiltfilt(sos, signal)
    return filtered


# Old method, doesn't work great for bandpass filters

# def _signal_filter_butterworth_ba(signal, sampling_rate=1000, lowcut=None, highcut=None, butterworth_order=2):
#    """Filter a signal using IIR Butterworth B/A method.
#    """
#    # Get coefficients
#    nyquist_freq = 0.5 * sampling_rate
#
#    if lowcut is not None:
#        lowcut = lowcut / nyquist_freq
#    if highcut is not None:
#        highcut = highcut / nyquist_freq
#
#    if lowcut is not None and highcut is not None:
#        b, a = scipy.signal.butter(butterworth_order, [lowcut, highcut], btype="bandpass", output='ba')
#    elif lowcut is not None:
#        b, a = scipy.signal.butter(butterworth_order, [lowcut], btype="highpass", output='ba')
#    elif highcut is not None:
#        b, a = scipy.signal.butter(butterworth_order, [highcut], btype="lowpass", output='ba')
#
#    try:
#        filtered = scipy.signal.filtfilt(b, a, signal, method="gust")
#    except ValueError:
#        filtered = scipy.signal.filtfilt(b, a, signal, method="pad")
#
#    return filtered


# =============================================================================
# Internals
# =============================================================================
def _signal_filter_sanitize(lowcut=None, highcut=None):
    if lowcut is not None and highcut is not None:
        if lowcut > highcut:
            filter_type = "bandstop"
        else:
            filter_type = "bandpass"
        freqs = [lowcut, highcut]
    elif lowcut is not None:
        freqs = [lowcut]
        filter_type = "highpass"
    elif highcut is not None:
        freqs = [highcut]
        filter_type = "lowpass"

    return freqs, filter_type





def _signal_filter_windowlength(window_length="default", sampling_rate=1000):
    if isinstance(window_length, str):
        window_length = int(np.round(sampling_rate/3))
        if (window_length % 2) == 0:
            window_length + 1
    return window_length



# -*- coding: utf-8 -*-
import mne

import pandas as pd
import numpy as np





def eeg_add_channel(raw, channel, channel_type=None, channel_name=None, sync_index_raw=0, sync_index_channel=0):
    """
    Add a channel to a mne's Raw m/eeg file. It will basically synchronize the channel to the eeg data following a particular index and add it.

    Parameters
    ----------
    raw : mne.io.Raw
        Raw EEG data from MNE.
    channel : list or array
        The signal to be added.
    channel_type : str
        Channel type. Currently supported fields are 'ecg', 'bio', 'stim', 'eog', 'misc', 'seeg', 'ecog', 'mag', 'eeg', 'ref_meg', 'grad', 'emg', 'hbr' or 'hbo'.
    channel_type : str
        Desired channel name.
    sync_index_raw, sync_index_channel : int or list
        An index (e.g., the onset of the same event marked in the same signal), in the raw data and in the channel to add, by which to align the two inputs. This can be used in case the EEG data and the channel to add do not have the same onsets and must be aligned through some common event.

    Returns
    ----------
    mne.io.Raw
        Raw data in FIF format.

    Example
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> # Let's say that the 42nd sample point in the EEG correspond to the 333rd point in the ECG
    >>> event_index_in_eeg = 42
    >>> event_index_in_ecg = 333
    >>>
    >>> raw = nk.eeg_add_channel(raw, ecg, sync_index_raw=event_index_in_eeg, sync_index_channel=event_index_in_ecg, channel_type="ecg")
    """
    if channel_name is None:
        if isinstance(channel, pd.Series):
            if channel.name is not None:
                channel_name = channel.name
            else:
                channel_name = "Added_Channel"
        else:
            channel_name = "Added_Channel"

    # Compute the distance between the two signals
    diff = sync_index_channel - sync_index_raw
    if diff > 0:
        channel = list(channel)[diff:len(channel)]
        channel = channel + [np.nan]*diff
    if diff < 0:
        channel = [np.nan]*abs(diff) + list(channel)

    # Adjust to raw size
    if len(channel) < len(raw):
        channel = list(channel) + [np.nan]*(len(raw)-len(channel))
    else:
        # Crop to fit the raw data
        channel = list(channel)[0:len(raw)]

    info = mne.create_info([channel_name], raw.info["sfreq"], ch_types=channel_type)
    channel = mne.io.RawArray([channel], info)

    raw.add_channels([channel], force_update_info=True)

    return(raw)





# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import scipy.signal


def signal_psd(signal, sampling_rate=1000, show=True):
    """Compute the Power Spectral Density (PSD).

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
        or "bandstop".
    sampling_rate : int
        The sampling frequency of the signal (in Hz, i.e., samples/second).
    show : bool
        If True, will return a plot. If False, will return the density values that can be plotted externally.

    See Also
    --------
    signal_filter

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the Power Spectrum values and a plot if
        `show` is True.

    Examples
    --------
    >>> import scipy.misc
    >>> import neurokit2 as nk
    >>>
    >>> signal = scipy.misc.electrocardiogram()
    >>> nk.signal_plot(signal)
    >>> nk.signal_psd(signal, sampling_rate=360, show=True)
    """
    frequency, power = scipy.signal.welch(signal,
                                          fs=sampling_rate,
                                          scaling='spectrum',
                                          detrend=False,
                                          average='median')

    data = pd.DataFrame({"Frequency": frequency,
                         "Power": power})

    if show is True:
        ax = data.plot(x="Frequency", y="Power", logy=True, title='Power Spectral Density (PSD)')
        ax.set(xlabel="Frequency (Hz)", ylabel="Spectrum")
        return ax
    else:
        return data



"""Submodule for NeuroKit."""

from .rsp_simulate import rsp_simulate
from .rsp_clean import rsp_clean
from .rsp_findpeaks import rsp_findpeaks
from .rsp_rate import rsp_rate
from .rsp_amplitude import rsp_amplitude
from .rsp_process import rsp_process
from .rsp_plot import rsp_plot



"""Top-level package for NeuroKit."""

# Info
__version__ = '0.0.6'
__citation__ = 'Not yet available.'
__bibtex__ = __citation__
__cite__ = __citation__


# Maintainer info
__author__ = """NeuroKit development team"""
__email__ = 'dom.makowski@gmail.com'



# Export content of submodules
from .misc import *
from .stats import *
from .complexity import *
from .signal import *
from .events import *
from .data import *

from .ecg import *
from .rsp import *
from .eda import *
from .emg import *
from .ppg import *
from .eeg import *



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from ..signal import signal_interpolate
from ..signal import signal_smooth


def rsp_rate(peaks, sampling_rate=1000, desired_length=None,
             method="khodadad2018"):
    """Compute respiration (RSP) rate.

    Compute respiration rate with the specified method.

    Parameters
    ----------
    peaks : list, array, DataFrame, Series or dict
        The samples at which the inhalation peaks occur. If a dict or a
        DataFrame is passed, it is assumed that these containers were obtained
        with `rsp_findpeaks()`.
    sampling_rate : int
        The sampling frequency of the signal that contains the peaks (in Hz,
        i.e., samples/second).
    desired_length : int
        By default, the returned respiration rate has the same number of
        elements as `peaks`. If set to an integer, the returned rate will be
        interpolated between `peaks` over `desired_length` samples. Has no
        effect if a DataFrame is passed in as the `peaks` argument.
    method : str
        The processing pipeline to apply. Can be one of "khodadad2018"
        (default) or "biosppy".

    Returns
    -------
    signals : DataFrame
        A DataFrame containing the respiration rate accessible with the key
        "RSP_Rate".

    See Also
    --------
    rsp_clean, rsp_findpeaks, rsp_amplitude, rsp_process, rsp_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> rsp = nk.rsp_simulate(duration=90, respiratory_rate=15)
    >>> cleaned = nk.rsp_clean(rsp, sampling_rate=1000)
    >>> signals, info = nk.rsp_findpeaks(cleaned)
    >>>
    >>> data = nk.rsp_rate(signals)
    >>> data["RSP_Signal"] = cleaned  # Add the signal back
    >>> data.plot(subplots=True)
    """
    # Sanity checks.
    if isinstance(peaks, dict):
        peaks = peaks["RSP_Peaks"]
    elif isinstance(peaks, pd.DataFrame):
        desired_length = len(peaks["RSP_Peaks"])
        peaks = np.where(peaks["RSP_Peaks"] == 1)[0]
    if len(peaks) <= 3:
        print("NeuroKit warning: rsp_rate(): too few peaks detected to "
              "compute the rate.")
        return

    # Find length of final signal to return.
    if desired_length is None:
        desired_length = len(peaks)

    # Calculate period in msec, based on horizontal peak to peak difference and
    # make sure that rate has the same number of elements as peaks (important
    # for interpolation later) by prepending the mean of all periods.
    period = np.ediff1d(peaks, to_begin=0) / sampling_rate
    period[0] = np.mean(period)
    rate = 60 / period

    # Preprocessing.
    rate, peaks = _rsp_rate_preprocessing(rate, peaks, method=method)

    # Interpolate rates to desired_length samples.
    rate = signal_interpolate(rate, x_axis=peaks,
                              desired_length=desired_length)

    signals = pd.DataFrame(rate, columns=["RSP_Rate"])

    return signals


# =============================================================================
# Internals
# =============================================================================
def _rsp_rate_preprocessing(rate, peaks, troughs=None, method="biosppy"):

    method = method.lower()  # remove capitalised letters
    if method == "biosppy":
        rate, peaks = _rsp_rate_outliers(rate, peaks, threshold_absolute=35)
        # Smooth with moving average
        rate = signal_smooth(signal=rate, kernel='boxcar', size=3)
    elif method == "khodadad2018":
        pass
    else:
        raise ValueError("NeuroKit error: rsp_rate(): 'method' should be "
                         "one of 'khodadad2018' or 'biosppy'.")

    return rate, peaks


def _rsp_rate_outliers(rate, peaks, troughs=None, threshold_absolute=35):

    if threshold_absolute is None:
        return rate, peaks

    # Enforce physiological limits.
    keep = np.nonzero(rate <= threshold_absolute)

    return rate[keep], peaks[keep]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import re


# Utilities
with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('NEWS.rst') as history_file:
    history = history_file.read()

def find_version():
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format("__version__"), open('neurokit2/__init__.py').read())
    return result.group(1)



# Dependencies
requirements = ['numpy', 'pandas', 'scipy', 'sklearn', 'mne', 'bioread', 'matplotlib']
setup_requirements = ['pytest-runner', 'numpy']
test_requirements = requirements + ['pytest', 'coverage', 'pyentrp', 'nolds', 'biosppy']

# Setup
setup(
    author="Dominique Makowski",
    author_email='dom.makowski@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="The Python Toolbox for Neurophysiological Signal Processing.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    include_package_data=True,
    keywords='neurokit2',
    name='neurokit2',
    setup_requires=setup_requirements,
    test_suite='pytest',
    tests_require=test_requirements,
    url='https://github.com/neuropsychology/NeuroKit',
    version=find_version(),
    zip_safe=False,
)



# -*- coding: utf-8 -*-
import numpy as np
import scipy.interpolate


def signal_interpolate(signal, desired_length=None, method="quadratic",
                       x_axis=None, new_x=None):
    """Interpolate a signal.

    Interpolate (fills the values between data points) a signal using different methods.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    desired_length : int
        The desired length of the signal.
    method : str
        Method of interpolation. Can be 'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'previous' or 'next'.  'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline interpolation of zeroth, first, second or third order; 'previous' and 'next' simply return the previous or next value of the point) or as an integer specifying the order of the spline interpolator to use.
    x_axis : list, array or Series
        An optional vector of same length as 'signal' corresponding to the x-axis.

    Returns
    -------
    array
        Vector of interpolated samples.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>> import matplotlib.pyplot as plt
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=10))
    >>> zero = nk.signal_interpolate(signal, desired_length=1000, method="zero")
    >>> linear = nk.signal_interpolate(signal, desired_length=1000, method="linear")
    >>> quadratic = nk.signal_interpolate(signal, desired_length=1000, method="quadratic")
    >>> cubic = nk.signal_interpolate(signal, desired_length=1000, method="cubic")
    >>> nearest = nk.signal_interpolate(signal, desired_length=1000, method="nearest")
    >>>
    >>> plt.plot(np.linspace(0, 1, num=len(zero)), zero, 'y',
                 np.linspace(0, 1, num=len(linear)), linear, 'r',
                 np.linspace(0, 1, num=len(quadratic)), quadratic, 'b',
                 np.linspace(0, 1, num=len(cubic)), cubic, 'g',
                 np.linspace(0, 1, num=len(nearest)), nearest, 'm',
                 np.linspace(0, 1, num=len(signal)), signal, 'ko')
    >>>
    >>> # Use x-axis end new x-axis
    >>> x_axis = np.linspace(start=10, stop=30, num=10)
    >>> signal = np.cos(x_axis)
    >>> new_x = np.linspace(start=0, stop=40, num=1000)
    >>> interpolated = nk.signal_interpolate(signal,
                                    desired_length=1000,
                                    x_axis=x_axis,
                                    new_x=new_x)
    >>> plt.plot(new_x, interpolated, '-',
                 x, signal, 'o')
    """
    # Sanity checks
    if desired_length is None:
        if new_x is not None:
            desired_length = new_x
        else:
            raise ValueError("NeuroKit error: signal_interpolate(): either 'desired_length' or 'new_x' must be provided.")
    if desired_length < len(signal):
        raise ValueError("NeuroKit error: signal_interpolate(): 'desired_length' cannot be lower than the length of the signal. You might be looking for 'signal_resample()'")

    # Create x axis
    if x_axis is None:
        x_axis = np.arange(0, len(signal))

    # Create interpolation function
    interpolation_function = scipy.interpolate.interp1d(
            x_axis,
            np.ravel(signal),
            kind=method,
            bounds_error=False,
            fill_value=([signal[0]], [signal[-1]]))

    if new_x is None:
        new_x = np.linspace(x_axis[0], x_axis[-1], num=desired_length)

    interpolated = interpolation_function(new_x)

    return(interpolated)



# -*- coding: utf-8 -*-
import numpy as np
import scipy.signal

from ..signal import signal_filter


def ecg_clean(ecg_signal, sampling_rate=1000, method="neurokit"):
    """Clean an ECG signal.

    Prepare a raw ECG signal for R-peak detection with the specified method.

    Parameters
    ----------
    ecg_signal : list, array or Series
        The raw ECG channel.
    sampling_rate : int
        The sampling frequency of `ecg_signal` (in Hz, i.e., samples/second).
        Defaults to 1000.
    method : str
        The processing pipeline to apply. Can be one of 'neurokit' (default)
        or 'biosppy'.

    Returns
    -------
    array
        Vector containing the cleaned ECG signal.

    See Also
    --------
    ecg_findpeaks, ecg_rate, ecg_process, ecg_plot

    Examples
    --------
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> ecg = nk.ecg_simulate(duration=10, sampling_rate=1000)
    >>> signals = pd.DataFrame({
            "ECG_Raw": ecg,
            "ECG_NeuroKit": nk.ecg_clean(ecg, sampling_rate=1000, method="neurokit"),
            "ECG_BioSPPy":nk.ecg_clean(ecg, sampling_rate=1000, method="biosppy")})
    >>> signals.plot()
    """
    method = method.lower()  # remove capitalised letters
    if method in ["neurokit", "nk"]:
        clean = _ecg_clean_nk(ecg_signal, sampling_rate=sampling_rate)
    elif method in ["biosppy"]:
        clean = _ecg_clean_biosppy(ecg_signal, sampling_rate=sampling_rate)
    else:
        raise ValueError("NeuroKit error: ecg_clean(): 'method' should be "
                         "one of 'neurokit' or 'biosppy'.")
    return clean





# =============================================================================
# Neurokit
# =============================================================================
def _ecg_clean_nk(signal, sampling_rate=1000):

    # Remove slow drift and dc offset with highpass Butterworth.
    clean = signal_filter(signal=signal,
                          sampling_rate=sampling_rate,
                          lowcut=0.5,
                          method="butterworth",
                          order=5)
    return clean



# =============================================================================
# Biosppy
# =============================================================================
def _ecg_clean_biosppy(signal, sampling_rate=1000):
    """
    adapted from https://github.com/PIA-Group/BioSPPy/blob/e65da30f6379852ecb98f8e2e0c9b4b5175416c3/biosppy/signals/ecg.py#L69
    """

    order = int(0.3 * sampling_rate)
    if order % 2 == 0:
        order += 1  # Enforce odd number

    # -> filter_signal()
    frequency = [3, 45]

    #   -> get_filter()
    #     -> _norm_freq()
    frequency = 2 * np.array(frequency) / sampling_rate  # Normalize frequency to Nyquist Frequency (Fs/2).

    #     -> get coeffs
    a = np.array([1])
    b = scipy.signal.firwin(numtaps=order, cutoff=frequency, pass_zero=False)

    # _filter_signal()
    filtered = scipy.signal.filtfilt(b, a, signal)

    return filtered



# -*- coding: utf-8 -*-
import numpy as np
import neurokit2 as nk
import matplotlib.pyplot as plt


def test_rsp_simulate():
    rsp1 = nk.rsp_simulate(duration=20, length=3000)
    assert len(rsp1) == 3000

    rsp2 = nk.rsp_simulate(duration=20, length=3000, respiratory_rate=80)
#    pd.DataFrame({"RSP1":rsp1, "RSP2":rsp2}).plot()
#    pd.DataFrame({"RSP1":rsp1, "RSP2":rsp2}).hist()
    assert (len(nk.signal_findpeaks(rsp1, height_min=0.2)[0]) <
            len(nk.signal_findpeaks(rsp2, height_min=0.2)[0]))

    rsp3 = nk.rsp_simulate(duration=20, length=3000, method="sinusoidal")
    rsp4 = nk.rsp_simulate(duration=20, length=3000, method="breathmetrics")
#    pd.DataFrame({"RSP3":rsp3, "RSP4":rsp4}).plot()
    assert (len(nk.signal_findpeaks(rsp3, height_min=0.2)[0]) >
            len(nk.signal_findpeaks(rsp4, height_min=0.2)[0]))


def test_rsp_clean():

    sampling_rate = 1000
    rsp = nk.rsp_simulate(duration=120, sampling_rate=sampling_rate,
                          respiratory_rate=15)

    khodadad2018 = nk.rsp_clean(rsp, sampling_rate=1000, method="khodadad2018")
    assert len(rsp) == len(khodadad2018)

    biosppy = nk.rsp_clean(rsp, sampling_rate=1000, method="biosppy")
    assert len(rsp) == len(biosppy)


    # Check if filter was applied.
    fft_raw = np.fft.rfft(rsp)
    fft_khodadad2018 = np.fft.rfft(khodadad2018)
    fft_biosppy = np.fft.rfft(biosppy)

    freqs = np.fft.rfftfreq(len(rsp), 1/sampling_rate)
#    assert np.sum(fft_raw[freqs > 2]) > np.sum(fft_khodadad2018[freqs > 2])
    assert np.sum(fft_raw[freqs > 2]) > np.sum(fft_biosppy[freqs > 2])
    assert np.sum(fft_khodadad2018[freqs > 2]) > np.sum(fft_biosppy[freqs > 2])

    # Check if detrending was applied.
    assert np.mean(rsp) > np.mean(khodadad2018)


def test_rsp_findpeaks():

    rsp = nk.rsp_simulate(duration=120, sampling_rate=1000,
                          respiratory_rate=15)
    rsp_cleaned = nk.rsp_clean(rsp, sampling_rate=1000)
    signals, info = nk.rsp_findpeaks(rsp_cleaned)
    assert signals.shape == (120000, 2)
    assert signals["RSP_Peaks"].sum() == 28
    assert signals["RSP_Troughs"].sum() == 28
    assert info["RSP_Peaks"].shape[0] == 28
    assert info["RSP_Troughs"].shape[0] == 28
    assert np.allclose(info["RSP_Peaks"].sum(), 1643787)
    assert np.allclose(info["RSP_Troughs"].sum(), 1586275)
    # Assert that extrema start with a trough and end with a peak.
    assert info["RSP_Peaks"][0] > info["RSP_Troughs"][0]
    assert info["RSP_Peaks"][-1] > info["RSP_Troughs"][-1]


def test_rsp_rate():

    rsp = nk.rsp_simulate(duration=120, sampling_rate=1000,
                          respiratory_rate=15, method="sinusoidal", noise=0)
    rsp_cleaned = nk.rsp_clean(rsp, sampling_rate=1000)
    signals, info = nk.rsp_findpeaks(rsp_cleaned)

    # Test with dictionary.
    test_length = 30
    rate = nk.rsp_rate(peaks=info, sampling_rate=1000,
                       desired_length=test_length)
    assert rate.shape == (test_length, 1)
    assert np.abs(rate["RSP_Rate"].mean() - 15) < 0.2

    # Test with DataFrame.
    rate = nk.rsp_rate(signals, sampling_rate=1000)
    assert rate.shape == (signals.shape[0], 1)
    assert np.abs(rate["RSP_Rate"].mean() - 15) < 0.2


def test_rsp_amplitude():

    rsp = nk.rsp_simulate(duration=120, sampling_rate=1000,
                          respiratory_rate=15, method="sinusoidal", noise=0)
    rsp_cleaned = nk.rsp_clean(rsp, sampling_rate=1000)
    signals, info = nk.rsp_findpeaks(rsp_cleaned)

    # Test with dictionary.
    test_length = 60
    amplitude = nk.rsp_amplitude(rsp_signal=rsp, extrema=info,
                                 desired_length=test_length)
    assert amplitude.shape == (test_length, 1)
    assert np.abs(amplitude["RSP_Amplitude"].mean() - 1) < 0.01

    # Test with DataFrame.
    amplitude = nk.rsp_amplitude(rsp_signal=rsp, extrema=signals)
    assert amplitude.shape == (rsp.size, 1)
    assert np.abs(amplitude["RSP_Amplitude"].mean() - 1) < 0.01


def test_rsp_process():

    rsp = nk.rsp_simulate(duration=120, sampling_rate=1000,
                          respiratory_rate=15)
    signals, info = nk.rsp_process(rsp, sampling_rate=1000)

    # Only check array dimensions since functions called by rsp_process have
    # already been unit tested.
    assert signals.shape == (120000, 6)
    assert np.array(["RSP_Raw",
                     "RSP_Clean",
                     "RSP_Peaks",
                     "RSP_Troughs",
                     "RSP_Rate",
                     "RSP_Amplitude"]) in signals.columns.values


def test_rsp_plot():

    rsp = nk.rsp_simulate(duration=120, sampling_rate=1000,
                          respiratory_rate=15)
    rsp_summary, _ = nk.rsp_process(rsp, sampling_rate=1000)
    nk.rsp_plot(rsp_summary)
    # This will identify the latest figure.
    fig = plt.gcf()
    assert len(fig.axes) == 3
    titles = ["Raw and Cleaned RSP",
              "Breathing Rate",
              "Breathing Amplitude"]
    for (ax, title) in zip(fig.get_axes(), titles):
        assert ax.get_title() == title
    plt.close(fig)



import numpy as np
import pandas as pd
import neurokit2 as nk

import scipy.stats

# =============================================================================
# EMG
# =============================================================================


def test_emg_simulate():

    emg1 = nk.emg_simulate(duration=20, length=5000, n_bursts=1)
    assert len(emg1) == 5000

    emg2 = nk.emg_simulate(duration=20, length=5000, n_bursts=15)
    assert scipy.stats.median_absolute_deviation(emg1) < scipy.stats.median_absolute_deviation(emg2)

    emg3 = nk.emg_simulate(duration=20, length=5000, n_bursts=1, duration_bursts=2.0)
#    pd.DataFrame({"EMG1":emg1, "EMG3": emg3}).plot()
    assert len(nk.signal_findpeaks(emg3, height_min=1.0)[0]) > len(nk.signal_findpeaks(emg1, height_min=1.0)[0])



# -*- coding: utf-8 -*-


def listify(**kwargs):
    """Transforms arguments into lists of the same length.

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> listify(a=3, b=[3, 5], c=[3])
    """
    args = kwargs
    maxi = 1

    # Find max length
    for key, value in args.items():
        if isinstance(value, str) is False:
            try:
                if len(value) > maxi:
                    maxi = len(value)
            except TypeError:
                pass

    # Transform to lists
    for key, value in args.items():
        if isinstance(value, list):
            args[key] = _multiply_list(value, maxi)
        else:
            args[key] = _multiply_list([value], maxi)

    return args





def _multiply_list(lst, length):
    q, r = divmod(length, len(lst))
    return q * lst + lst[:r]



# -*- coding: utf-8 -*-
import pandas as pd
from .rsp_clean import rsp_clean
from .rsp_findpeaks import rsp_findpeaks
from .rsp_rate import rsp_rate
from .rsp_amplitude import rsp_amplitude


def rsp_process(rsp_signal, sampling_rate=1000, method="khodadad2018"):
    """Process a respiration (RSP) signal.

    Convenience function that automatically processes a respiration signal with
    one of the following methods:

    - `Khodadad et al. (2018)
    <https://iopscience.iop.org/article/10.1088/1361-6579/aad7e6/meta>`_
    - `BioSPPy
    <https://github.com/PIA-Group/BioSPPy/blob/master/biosppy/signals/resp.py>`_

    Parameters
    ----------
    rsp_signal : list, array or Series
        The raw respiration channel (as measured, for instance, by a
        respiration belt).
    sampling_rate : int
        The sampling frequency of `rsp_signal` (in Hz, i.e., samples/second).
    method : str
        The processing pipeline to apply. Can be one of "khodadad2018"
        (default) or "biosppy".

    Returns
    -------
    signals : DataFrame
        A DataFrame f same length as `rsp_signal` containing the following
        columns:

        - *"RSP_Raw"*: the raw signal.
        - *"RSP_Clean"*: the cleaned signal.
        - *"RSP_Peaks"*: the inhalation peaks marked as "1" in a list of zeros.
        - *"RSP_Troughs"*: the exhalation troughs marked as "1" in a list of
                            zeros.
        - *"RSP_Rate"*: breathing rate interpolated between inhalation peaks.
        - *"RSP_Amplitude"*: breathing amplitude interpolated between
                                inhalation peaks.
    info : dict
        A dictionary containing the samples at which inhalation peaks and
        exhalation troughs occur, accessible with the keys "RSP_Peaks", and
        "RSP_Troughs", respectively.

    See Also
    --------
    rsp_clean, rsp_findpeaks, rsp_rate, rsp_amplitude, rsp_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> rsp = nk.rsp_simulate(duration=90, respiratory_rate=15)
    >>> signals, info = nk.rsp_process(rsp, sampling_rate=1000)
    >>> nk.rsp_plot(signals)
    """
    rsp_cleaned = rsp_clean(rsp_signal, sampling_rate=sampling_rate,
                            method=method)

    extrema_signal, info = rsp_findpeaks(rsp_cleaned, method=method,
                                         outlier_threshold=0.3)

    rate = rsp_rate(extrema_signal, sampling_rate=sampling_rate, method=method)

    amplitude = rsp_amplitude(rsp_signal, extrema_signal)

    signals = pd.DataFrame({"RSP_Raw": rsp_signal,
                            "RSP_Clean": rsp_cleaned})
    signals = pd.concat([signals, extrema_signal, rate, amplitude], axis=1)

    return signals, info



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np




def entropy_shannon(signal):
    """Compute the Shannon entrop (SE).

    Entropy is a measure of unpredictability of the state, or equivalently, of its average information content. Shannon entropy (SE) is one of the first and most basic measure of entropy and a foundational concept of information theory. Shannon’s entropy quantifies the amount of information in a variable.


    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.


    Returns
    ----------
    float
        The Shannon entropy as float value.

    See Also
    --------
    entropy_approximate, entropy_sample, entropy_fuzzy

    Examples
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=30, num=100))
    >>> nk.entropy_shannon(signal)
    6.6438561897747395


    References
    -----------
    - `pyEntropy` <https://github.com/nikdon/pyEntropy>`_
    - `EntroPy` <https://github.com/raphaelvallat/entropy>`_
    - `nolds` <https://github.com/CSchoel/nolds>`_
    """
    # Check if string
    if not isinstance(signal, str):
        signal = list(signal)

    signal = np.array(signal)

    # Create a frequency data
    data_set = list(set(signal))
    freq_list = []
    for entry in data_set:
        counter = 0.
        for i in signal:
            if i == entry:
                counter += 1
        freq_list.append(float(counter) / len(signal))

    # Shannon entropy
    shannon_entropy = 0.0
    for freq in freq_list:
        shannon_entropy += freq * np.log2(freq)
    shannon_entropy = -shannon_entropy

    return shannon_entropy
